var blackboard = document.getElementById("blackboard");
var tacto;// Constructed later in init()

function $(id) { return document.getElementById(id); }

function message(text)
{
    var container = $("message");
    if (container.firstChild)
    {
	container.insertBefore(document.createElement("br"),
			       container.firstChild);
	container.insertBefore(document.createTextNode(text),
			       container.firstChild);
    }
    else
    {
	container.appendChild(document.createTextNode(text));
    }
}

// Keyboard sound:
var keyboard_click_active = false;
var keyboard_click = $("keyboard-click");
if (keyboard_click_active)
{
    keyboard_click.load();
    keyboard_click.autobuffer = true;
}

var display_timer;
function palette_click(event)
{
    tacto.standard_event(event);
    var prevent_default = true;
    if (event.target.nodeName.toLowerCase() == "input")
    {
	if (keyboard_click_active)
	{
	    (new Audio(keyboard_click.src)).play();
	}
	switch (event.target.name)
	{
	case "up":
	case "left":
	case "down":
	case "right":
	case "backspace":
	case "baseline":
	    event.preventDefault();
	    tacto.do_command(event.target.name);
	    blackboard.focus();
	    break;
	case "keyboard":
	    event.preventDefault();
	    $("keyboard-input").focus();
	    break;
	case "keyboard-input":
	    break;
	case "get-tree":
	    event.preventDefault();
	    display_formula();
	    blackboard.focus();
	    break;
	case "toggle-palettes":
	    event.preventDefault();
	    var node = $("palettes");
	    if (node.style.height == "auto" || !node.style.height)
	    {
		node.style.height = "2em";
	    }
	    else
	    {
		node.style.height = "auto";
	    };
	    blackboard.focus();
	    break;
	default:
	    event.preventDefault();
	    tacto.insert(event.target.value);
	    blackboard.focus();
	}
    }
    tacto.show_cursor(true);
}

function display_formula_error(message)
{
    var error = document.createElement("div");
    // This is for Internet Explorer:
    if ("textContent" in error) error.textContent = message;
    else                        error.innerText   = message;
    
    return error;
}
var previous_voice;
function display_formula()
{
    var formula_display = $("formula-display");
    while (formula_display.firstChild)
	formula_display.removeChild(formula_display.firstChild);
    var div = document.createElement("div");
    var formats = ["pmml",
		   "bst",
		   "openmath",
		   "text"];
    if ($("voice-active").checked) formats.push("voice");
    
    var expression = tacto.get_expression(formats);
    var result, i;
    result = expression.pmml;
    if (typeof result == "string")
    {
	div.appendChild(display_formula_error("MathML: " + result));
    }
    else
    {
	for (i = 0; i < result.length; ++i) div.appendChild(result[i]);
    }
    div.appendChild(document.createElement("br"));
    div.appendChild(document.createElement("br"));
    result = expression.bst;
    if (typeof result == "string")
    {
	div.appendChild(display_formula_error("BST: " + result));
    }
    else
    {
	div.appendChild(document.createTextNode(result));
    }
    div.appendChild(document.createElement("br"));
    div.appendChild(document.createElement("br"));
    result = expression.openmath;
    if (typeof result == "string")
    {
	div.appendChild(display_formula_error("OpenMath: " + result));
    }
    else
    {
	var serializer = new XMLSerializer();
	var sentences = expression.text;
	if (typeof sentences == "string")
	{
	    sentences = ["Error when building English form: "
			 + sentences];
	}
	for (i = 0; i < result.length; ++i)
        {
            if (sentences && i < sentences.length)
	    {
		div.appendChild(document.createTextNode(sentences[i]));
		div.appendChild(document.createElement("br"));
		div.appendChild(document.createElement("br"));
	    }
            div.appendChild(document.createTextNode
			    (serializer.serializeToString(result[i]))
			   );
	    div.appendChild(document.createElement("br"));
	    div.appendChild(document.createElement("br"));
        }
    }
    if ("voice" in expression)
    {
	result = expression.voice;
	if (typeof result == "string")
	{
	    message("Voice error: " + result);
	}
	else
	{
	    if (previous_voice) previous_voice.stop();
	    result.speak();
	    previous_voice = result;
	}
    }
    
    formula_display.appendChild(div);
}

function adjust_size()
{
    var available_height;
    if ("innerHeight" in window)
    {
	available_height = window.innerHeight;
    }
    else
    {
	// Fallback for Internet Explorer
	available_height = document.documentElement.clientHeight;
    }
    available_height -= blackboard.offsetTop;
    blackboard.style.height = available_height + "px";
    // This fails in Opera 9
    var palettes = $("palettes");
    if (!palettes) return;
    
    var font_size = null;
    if (palettes.offsetWidth > palettes.offsetHeight
	&& palettes.offsetHeight > available_height
	|| palettes.offsetHeight > palettes.offsetWidth
	&& palettes.offsetWidth > blackboard.offsetWidth/2)
    {
	font_size = "3mm";
    }
    var node = palettes.firstChild;
    while (node)
    {
	if (node.nodeType == node.ELEMENT_NODE)
	{
	    if (font_size) node.style.fontSize = font_size;
	}
	node = node.nextSibling;
    }
    if (font_size) $("palette-mini").style.fontSize = font_size;
    window.onresize = function (event)
    { window.setTimeout("adjust_size()", 100); };
}

function set_theme(name)
{
    blackboard.className = name;
    tacto.font_changed();
}

function formula_changed(formula)
{
    if (display_timer) window.clearTimeout(display_timer);
    display_timer = window.setTimeout(function () {
	display_formula();
	display_timer = null;
    }, 100);
}

function init()
{
    tacto = new Tacto_editor(blackboard);
    adjust_size();
    tacto.attach_text_input($("keyboard-input"));
    tacto.onchange = formula_changed;
    tacto.options.lock_rotation   = $("lock_rotation").checked;
    tacto.options.live_preview    = $("live_preview").checked;
    if (navigator.platform == 'iPad' || navigator.platform == 'iPod')
    {
	// In iOS v4.2 HTML5 audio playing from JavaScript
	// does not work at all, because Apple disabled it:
	// http://stackoverflow.com/questions/3009888/
	// autoplay-audio-files-on-an-ipad-with-html5
	$("voice-active").disabled = true;
	// Update: This is linked with the pop-up blocker in Safari.
	// If pop-ups are allowed, automatic audio playing works too,
	// but the quality is still atrocious so we disable it in
	// any case.
    }
    
    // Initial formula example:
    var x = (blackboard.clientWidth -
	     document.getElementById("palettes").clientWidth
	    ) / 2 - 100;
    if (x < 10) x = 10;
    tacto.move_to(x, blackboard.clientHeight / 2);
    tacto.insert("3(x+1)^2_<yz");
}
